use std::path::Path;
use usvg::tiny_skia_path::PathSegment;

pub fn load<P :AsRef<Path>>(file:P) -> Result<String,()>
{
    let mut output = String::from("new GameImage([");

    let svg_data = std::fs::read(file).map_err(|_| ())?;

    let opt = usvg::Options::default();
    let rtree = usvg::Tree::from_data(&svg_data, &opt).map_err(|_| ())?;

    let bounds = rtree.size();
    let origin = [
        bounds.width() / 2.,
        bounds.height() / 2.,
    ];
    let scale = 1. / bounds.width().max(bounds.height());

    for node in rtree.root().children() {
        output += &load_node(node, origin, scale)?;
    }

    output += "])";
    Ok(output)
}

fn load_node(node:&usvg::Node, origin:[f32; 2], scale:f32) -> Result<String,()>
{
    let mut output = String::new();

    match &node {
        usvg::Node::Group(group) => {
            for node in group.children() {
                output += &load_node(node, origin, scale)?;
            }
        }
        usvg::Node::Path(path) => {
            output += "[";
            output += &if let Some(fill) = path.fill() {
                match fill.paint() {
                    usvg::Paint::Color(color) => {
                        format!("\"#{:02x}{:02x}{:02x}\"", color.red, color.green, color.blue)
                    }
                    _ => {
                        String::from("\"#000000\"")
                    }
                }
            } else {
                String::from("\"#000000\"")
            };
            output += ",[";

            for ref segment in path.data().segments() {
                match segment {
                    PathSegment::MoveTo(point) => {
                        output += &format!("[0, [{:.4},{:.4}]],",
                            (point.x - origin[0]) * scale,
                            (point.y - origin[1]) * scale,
                        );
                    }
                    PathSegment::LineTo(point) => {
                        output += &format!("[1, [{:.4},{:.4}]],",
                            (point.x - origin[0]) * scale,
                            (point.y - origin[1]) * scale,
                        );
                    }
                    PathSegment::CubicTo(a, b, c) => {
                        output += &format!("[2, [{:.4},{:.4},{:.4},{:.4},{:.4},{:.4},]],",
                            (a.x - origin[0]) * scale,
                            (a.y - origin[1]) * scale,
                            (b.x - origin[0]) * scale,
                            (b.y - origin[1]) * scale,
                            (c.x - origin[0]) * scale,
                            (c.y - origin[1]) * scale,
                        );
                    }
                    PathSegment::Close => {
                        output += "[3],";
                    }
                    _ => { }
                }
            }

            output += "]],";
        }
        _ => { }
    }

    Ok(output)
}

